/* vim: set shiftwidth=2 tabstop=8 autoindent cindent expandtab: */
/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

#include "nsISupports.idl"

/**
 * Transition events are defined in:
 * http://www.w3.org/TR/css3-transitions/#transition-events-
 * http://dev.w3.org/csswg/css3-transitions/#transition-events-
 */

[builtinclass, uuid(ee3499bf-0f14-4bb6-829c-19ad24fd4a85)]
interface nsIDOMTransitionEvent : nsISupports {
  readonly attribute DOMString           propertyName;
  readonly attribute float               elapsedTime;
  readonly attribute DOMString           pseudoElement;
};
